# -*- coding: utf-8 -*-

"""XML Validation.

XML Schema Definition (XSD):
    https://www.w3.org/XML/Schema

Document Type Definition (DTD):
    https://en.wikipedia.org/wiki/Document_type_definition

RELAX NG:
    https://relaxng.org/

Validation with lxml:
    https://lxml.de/validation.html
"""


from logging import getLogger
import sys

# pylint: disable=no-member
# lxml ElementTree <https://lxml.de/>
from lxml import etree

# Import my own modules.
from .etree import build_etree


# Module logging initialisation.
logger = getLogger(__name__)


def build_xml_schema(xsd_file):
    """Parse an XSD file into an XMLSchema validator.

    xsd_file -- XSD file (XML schema file)

    Return XMLSchema validator (lxml.etree.XMLSchema) on success.
    Return None on error.

    The lxml.etree.XMLSchema class:
        https://lxml.de/validation.html#xmlschema
        https://lxml.de/api/lxml.etree.XMLSchema-class.html
    """
    xsd_etree = build_etree(xsd_file, lenient=False)
    if not xsd_etree:
        return None

    try:
        validator = etree.XMLSchema(xsd_etree)
    # Catch XSD parse errors.
    except etree.XMLSchemaParseError as inst:
        if inst.error_log.last_error.line == 0:
            # The XML document '%s' is not a schema document.
            logger.error(inst)
            return None
        logger.error("XML file '%s' is not a valid XSD file", xsd_file)
        # Lines with XSD parse errors are logged as warnings.
        for e in inst.error_log:
            # E.g. e.level_name: "ERROR", e.domain_name: "SCHEMASP",
            # e.type_name: "SCHEMAP_S4S_ATTR_INVALID_VALUE".
            logger.warning("line %i, column %i: %s", e.line, e.column, e.message)
        return None
    else:
        return validator


def build_dtd(dtd_file):
    """Parse a DTD file into a DTD validator.

    dtd_file -- DTD file

    Return DTD validator (lxml.etree.DTD) on success.
    Return None on error.

    The lxml.etree.DTD class:
        https://lxml.de/validation.html#id1
        https://lxml.de/api/lxml.etree.DTD-class.html
    """
    try:
        validator = etree.DTD(file=dtd_file)
    # Catch DTD parse errors.
    except etree.DTDParseError as inst:
        if inst.error_log.last_error.line == 0:
            # error parsing DTD
            logger.error(inst)
            # "failed to load external entity" when file does not exist.
            logger.error(inst.error_log.last_error.message)
            return None
        logger.error("'%s' is not a valid DTD file", dtd_file)
        # Lines with DTD parse errors are logged as warnings.
        for e in inst.error_log:
            # E.g. e.level_name: "FATAL", e.domain_name: "PARSER",
            # e.type_name: "ERR_EXT_SUBSET_NOT_FINISHED"
            logger.warning("line %i, column %i: %s", e.line, e.column, e.message)
        return None
    else:
        return validator


def xml_validator(xml_source, validator, lenient=True):
    """Validate an XML source against an XSD, DTD or RELAX NG validator.

    xml_source -- XML file, file-like object or URL
    validator -- XMLSchema, DTD validator or RELAX NG validator
    lenient -- log XML (validation) errors as warnings

    Return a tuple with the validation result (True/False) and the status string.
    """
    el_tree = build_etree(xml_source, lenient=lenient)
    if not el_tree:
        return (False, "Not an XML source")

    if xml_source in ('-', sys.stdin):
        # <stdin>.
        source_name = sys.stdin.name
    else:
        source_name = xml_source
    if validator.validate(el_tree):
        logger.info("XML source '%s' validates", source_name)
        return (True, "XML source validates")

    if lenient:
        val_logger = logger.warning
    else:
        val_logger = logger.error
    val_logger("XML source '%s' does not validate", source_name)
    # Lines with XML validation errors.
    for e in validator.error_log:
        # E.g. DTD e.level_name: "ERROR", e.domain_name: "VALID",
        # e.type_name: "DTD_UNKNOWN_ELEM".
        # E.g. XSD e.level_name: "ERROR", e.domain_name: "SCHEMASV",
        # e.type_name: "SCHEMAV_CVC_ELT_1".
        val_logger("line %i, column %i: %s", e.line, e.column, e.message)
    # Return the status string: first validation error.
    e = validator.error_log[0]
    return (False, "line %i, column %i: %s" % (e.line, e.column, e.message))


def build_relaxng(relaxng_file):
    """Parse a RELAX NG file into a RELAX NG validator.

    relaxng_file -- RELAX NG file

    Return RelaxNG validator (lxml.etree.RelaxNG) on success.
    Return None on error.

    The lxml.etree.RelaxNG class:
        https://lxml.de/validation.html#relaxng
        https://lxml.de/api/lxml.etree.RelaxNG-class.html
    """
    relaxng_etree = build_etree(relaxng_file, lenient=False)
    if not relaxng_etree:
        return None

    try:
        validator = etree.RelaxNG(relaxng_etree)
    # Catch RELAX NG parse errors.
    except etree.RelaxNGParseError as inst:
        logger.error("XML file '%s' is not a valid RELAX NG file", relaxng_file)
        if inst.error_log.last_error.line == -1:
            # xmlRelaxNGParse: schemas is empty.
            logger.error(inst)
            return None
        # Lines with RELAX NG parse errors are logged as warnings.
        for e in inst.error_log:
            # E.g. e.level_name: "ERROR", e.domain_name: "RELAXNGP",
            # e.type_name: "RNGP_UNKNOWN_CONSTRUCT".
            logger.warning("line %i, column %i: %s", e.line, e.column, e.message)
        return None
    else:
        return validator


def validate_xml(xml_source, validator, lenient=True, silent=False):
    """Validate an XML source against an XSD, DTD or RELAX NG validator.

    xml_source -- XML file, file-like object or URL
    validator -- XMLSchema, DTD validator or RELAX NG validator
    lenient -- log XML (validation) errors as warnings
    silent -- no logging

    Return the validation result (True/False).
    """
    el_tree = build_etree(xml_source, lenient=lenient, silent=silent)
    if not el_tree:
        return False

    if xml_source in ('-', sys.stdin):
        # <stdin>.
        source_name = sys.stdin.name
    else:
        source_name = xml_source

    if validator.validate(el_tree):
        if not silent:
            logger.info("XML source '%s' validates", source_name)
        return True
    if silent:
        return False

    # Log validation errors.
    if lenient:
        val_logger = logger.warning
    else:
        val_logger = logger.error
    val_logger("XML source '%s' does not validate", source_name)
    # Lines with XML validation errors.
    for e in validator.error_log:
        # E.g. DTD e.level_name: "ERROR", e.domain_name: "VALID",
        # e.type_name: "DTD_UNKNOWN_ELEM".
        # E.g. XSD e.level_name: "ERROR", e.domain_name: "SCHEMASV",
        # e.type_name: "SCHEMAV_CVC_ELT_1".
        val_logger("line %i, column %i: %s", e.line, e.column, e.message)
    return False



# -*- coding: utf-8 -*-

"""Select nodes in an XML source with an XPath expression."""


from __future__ import print_function

from argparse import ArgumentParser
import sys
import errno

# pylint: disable=no-name-in-module
# lxml ElementTree <https://lxml.de/>
from lxml.etree import XPathEvalError, XMLParser
from lxml.etree import iselement, PI, Comment

# Import my own modules.
from .. import __version__
from ..log import setup_logger_console
from ..etree import build_etree
from ..xpath import build_xpath, etree_xpath, namespaces
from ..ppxml import prettyprint


def parse_cl():
    """Parse the command line for options, XPath expression and XML sources."""
    parser = ArgumentParser(
        description=__doc__)
    parser.add_argument(
        "-V", "--version", action="version",
        version="%(prog)s " + __version__)
    parser.add_argument("xpath_expr", help="XPath expression")
    parser.add_argument(
        "xml_sources", nargs='*',
        metavar='xml_source', help="XML source (file, <stdin>, http://...)")
    parser.add_argument(
        "-e", "--exslt",
        action="store_true", default=False, dest="exslt",
        help="add EXSLT XML namespace prefixes")
    parser.add_argument(
        "-d", "--default-prefix",
        action="store", default="d", dest="default_ns_prefix",
        help="set the prefix for the default namespace in XPath [default: '%(default)s']")
    parser.add_argument(
        "-r", "--result-xpath",
        action="store_true", default=False, dest="result_xpath",
        help="print the XPath expression of the result element (or its parent)")
    parser.add_argument(
        "-p", "--pretty-element",
        action="store_true", default=False, dest="pretty_element",
        help="pretty print the result element")
    parser.add_argument(
        "-m", "--method",
        action="store_true", default=False, dest="lxml_method",
        help="use ElementTree.xpath method instead of XPath class")
    file_group = parser.add_mutually_exclusive_group(required=False)
    file_group.add_argument(
        "-f", "-l", "--files-with-hits",
        action="store_true", default=False, dest="files_with_hits",
        help="only the names of files with a non-false and non-NaN result " +
        "are written to standard output")
    file_group.add_argument(
        "-F", "-L", "--files-without-hits",
        action="store_true", default=False, dest="files_without_hits",
        help="only the names of files with a false or NaN result, " +
        "or without any results are written to standard output")
    parser.add_argument(
        "-q", "--quiet",
        action="store_false", default=True, dest="verbose",
        help="don't print the XML namespace list")

    return parser.parse_args()


def xpath_class(el_tree, xpath_exp, ns_map):
    """XPath with lxml.etree.XPath class."""
    xpath_obj = build_xpath(xpath_exp, ns_map)
    if not xpath_obj:
        return None
    return etree_xpath(el_tree, xpath_obj)


def eltree_xpath(el_tree, xpath_exp, ns_map):
    """XPath with lxml.etree.ElementTree.xpath method."""
    try:
        xp_result = el_tree.xpath(xpath_exp, namespaces=ns_map)
    except XPathEvalError as e:
        sys.stderr.write(
            "XPath '%s' evaluation error: %s\n" % (xpath_exp, e))
        return None
    # EXSLT function call errors (re:test positional arguments).
    except TypeError as e:
        sys.stderr.write("XPath '%s' type error: %s\n" % (xpath_exp, e))
        return None
    else:
        return xp_result


def xp_prepare(args):
    """Return XPath function and XML parser.

    args -- Command-line arguments
    """
    # ElementTree.xpath method or XPath class (default).
    if args.lxml_method:
        xpath_fn = eltree_xpath
    else:
        xpath_fn = xpath_class

    # Initialise XML parser.
    if args.pretty_element:
        # Pretty print preparation (removes white space text nodes!).
        xml_parser = XMLParser(remove_blank_text=True)
    else:
        xml_parser = XMLParser()

    return xpath_fn, xml_parser


def print_xmlns(ns_map, root):
    """Print XML namespaces."""
    if None in root.nsmap:
        print("Default XML namespace URI: %s" % root.nsmap[None])
    if ns_map:
        # Print all XML namespaces -- prefix: namespace URI.
        print("XML namespaces:")
        for key in ns_map:
            print("%8s: %s" % (key, ns_map[key]))


def element_repr(node, content=True):
    """Return element representation (UTF-8 Unicode).

    node -- lxml.etree._Element instance -- iselement(node)

    Node types:
     * element node -- /path/el, //*
     * comment node -- comment()
     * processing instruction node -- processing-instruction()
    """
    if not content:
        # Return lxml.etree._Element representation without content.
        if node.tag is PI:
            elem_str = node.tag(node.target)
        elif node.tag is Comment:
            elem_str = node.tag(" comment ")
        else:
            elem_str = "<%s>" % node.tag
        return elem_str

    # node.tag is lxml.etree.PI (is lxml.etree.ProcessingInstruction).
    if node.tag is PI:
        # Processing instruction node - node.target -- node.tag(): <? ?>
        return "%s value: '%s'" % (node.tag(node.target), node.text)

    # node.tag is lxml.etree.Comment.
    if node.tag is Comment:
        # Comment node - node.tag(): <!-- -->
        return node.tag(node.text)

    # node.tag: string.
    if node.text:
        if node.text.isspace():
            elem_str = "<%s> contains whitespace" % node.tag
        elif not isinstance(node.text, str):
            # Python 2 Unicode naar Bytestring.
            elem_str = "<%s> contains '%s'" % (node.tag, node.text.encode("utf-8"))
        else:
            # node.text is a Python string.
            elem_str = "<%s> contains '%s'" % (node.tag, node.text)
        return elem_str
    return "<%s> is empty" % node.tag


def print_elem(node, pretty=False, xpath_exp=None):
    """Print element (UTF-8 Unicode).

    node -- lxml.etree._Element instance.
            element, comment or processing instruction node; see element_repr()
    pretty -- True: use prettyprint() to print an element.
              False: use element_repr().
    xpath_exp -- optional XPath expression
    """
    if pretty:
        if xpath_exp:
            print("XPath %s (line %d):" % (xpath_exp, node.sourceline))
        else:
            print("line %d:" % node.sourceline)
        prettyprint(node, xml_declaration=False)
    else:
        if xpath_exp:
            print("XPath %s (line %d):" % (xpath_exp, node.sourceline))
            print("   %s" % element_repr(node))
        else:
            print("line %4d:   %s" % (node.sourceline, element_repr(node)))


def smart_with_parent(smart_string):
    """Return lxml 'smart' string representation (UTF-8) with parent relation.

    lxml 'smart' string is a text node (atomic value) or an attribute node:
     * text node (tail, entity): contains text; never empty
     * attribute node: contains the value of the attribute
    """
    smart_repr = None
    parent_rel = None

    # ATTRIBUTE node -- @ -- .is_attribute
    if smart_string.is_attribute:
        parent_rel = "of"
        smart_repr = "@%s = '%s'" % (smart_string.attrname, smart_string)
    # TEXT node -- text() -- .is_text
    elif smart_string.is_text:
        parent_rel = "in"
        # Python str.isspace()
        if smart_string.isspace():
            smart_repr = "whitespace"
        else:
            smart_repr = "'%s'" % smart_string
    # TAIL node -- text() -- .is_tail
    elif smart_string.is_tail:
        parent_rel = "after"
        if smart_string.isspace():
            smart_repr = "tail whitespace"
        else:
            smart_repr = "tail '%s'" % smart_string

    return (smart_repr, parent_rel)


def print_smart_string(smart_string, el_tree, args):
    """Print lxml 'smart' string with parent element tag.

    smart_string -- XPath string result that provides a getparent() method:
     * string: lxml.etree._ElementStringResult
     * Unicode: lxml.etree._ElementUnicodeResult
    el_tree -- ElementTree (lxml.etree._ElementTree)
    args -- Command-line arguments
    """
    # Parent element.
    par_el = smart_string.getparent()
    # string() and concat() results do not have an origin.
    if par_el is None:
        print("XPath string: '%s'" % smart_string)
        return
    # Parent is an lxml.etree._Element instance.
    par_el_str = element_repr(par_el, content=False)

    # Print 'smart' string.
    smart_repr, parent_rel = smart_with_parent(smart_string)
    if smart_repr:
        if args.result_xpath:
            # Print the absolute XPath expression of the parent element.
            print("line %d, parent XPath %s" % (
                par_el.sourceline, el_tree.getpath(par_el)))
            print("   %s %s %s" % (smart_repr, parent_rel, par_el_str))
        else:
            print("line %4d:   %s %s %s" % (
                par_el.sourceline, smart_repr, parent_rel, par_el_str))
    else:
        print("**smart string DEBUG fallback**")
        print_elem(par_el, pretty=args.pretty_element)


def print_result_list(result_list, el_tree, args):
    """Print all nodes from the list of XPath results.

    result_list -- XPath result list
    el_tree -- ElementTree (lxml.etree._ElementTree)
    args -- Command-line arguments
    """
    # All nodes -- //node()
    for node in result_list:
        if iselement(node):
            if args.result_xpath:
                print_elem(
                    node,
                    pretty=args.pretty_element,
                    xpath_exp=el_tree.getpath(node))
            else:
                print_elem(node, pretty=args.pretty_element)

        # Smart string -- .getparent() | attribute, entity, text (atomic value).
        elif hasattr(node, "getparent"):
            print_smart_string(node, el_tree, args)

        # Namespaces -- namespace::
        elif isinstance(node, tuple):
            # No line number.
            print("prefix: %-8s URI: %s" % node)

        # ?
        else:
            print("**DEBUG fallback**")
            print(type(node))
            print(node)


def print_result_header(source_name, xp_result):
    """Print header with XPath result summary."""
    print("%s:" % source_name, end=" ")
    # XPath result summary.
    if isinstance(xp_result, list):
        list_result = xp_result
        xp_r_len = len(list_result)
    else:
        # String, number, boolean.
        list_result = [xp_result]
        xp_r_len = 1
    if xp_r_len == 0:
        print("no results.")
    elif xp_r_len == 1:
        if isinstance(list_result[0], tuple):
            print("1 XML namespace result.")
        else:
            print("1 result.")
    else:
        if isinstance(list_result[0], tuple):
            print("%d XML namespace results." % xp_r_len)
        else:
            print("%d results." % xp_r_len)


def print_xp_result(xp_result, el_tree, ns_map, args):
    """Print XPath results.

    xp_result -- XPath result
    el_tree -- ElementTree (lxml.etree._ElementTree)
    ns_map -- XML namespaces (xmlns) 'prefix: URI' dict
    args -- Command-line arguments

    Prints:
     * XML namespaces (if there are any)
     * XPath result(s)

    XPath return values:
        https://lxml.de/xpathxslt.html#xpath-return-values
    """
    if args.verbose:
        print_xmlns(ns_map, el_tree.getroot())

    # STRING - string (basestring) - smart string | Namespace URI.
    try:
        basestring
    except NameError:
        # Python 3 -- pylint: disable=redefined-builtin
        basestring = (str, bytes)
    # 'lxml.etree._ElementStringResult'
    if isinstance(xp_result, basestring):
        print_smart_string(xp_result, el_tree, args)

    # LIST - list - node-set.
    elif isinstance(xp_result, list):
        try:
            # List can be empty.
            print_result_list(xp_result, el_tree, args)
        except IOError as e:
            # Python 2: catch 'IOError: [Errno 32] Broken pipe'.
            if e.errno != errno.EPIPE:
                sys.stderr.write("IOError: %s [%s]\n" % (e.strerror, e.errno))

    # FLOAT - float.
    elif hasattr(xp_result, "is_integer"):
        # pylint: disable=comparison-with-itself ## NaN elif.
        if xp_result.is_integer():
            print("XPath number: %i" % xp_result)
        # float('nan') != float('nan') -- IEEE 754.
        elif xp_result != xp_result:
            print("XPath result: NaN (not a number)")
        # float.
        else:
            print("XPath number: %s" % xp_result)

    # BOOLEAN - bool - boolean.
    elif isinstance(xp_result, bool):
        print("XPath test: %s" % xp_result)
    else:
        print("Unknown XPath result: %s" % xp_result)


def xpath_on_xml(xml_source, parser, xpath_fn, args):
    """Apply XPath expression to XML source.

    xml_source -- XML file, file-like object or URL
    parser -- XML parser (lxml.etree.XMLParser)
    xpath_fn -- ElementTree.xpath method or XPath class
    args -- Command-line arguments
    """
    # ElementTree (lxml.etree._ElementTree).
    el_tree = build_etree(xml_source, parser=parser, lenient=False)
    if el_tree is None:
        return False

    # XML namespaces.
    ns_map = namespaces(el_tree, args.exslt, args.default_ns_prefix)
    # XPath expression on ElementTree.
    xp_result = xpath_fn(el_tree, args.xpath_expr, ns_map)
    if xp_result is None:
        return False

    # Printable name for sys.stdin.
    if xml_source in ('-', sys.stdin):
        source_name = sys.stdin.name
    else:
        source_name = xml_source

    # XML sources names (--files-with-results/--files-without-results).
    if args.files_with_hits or args.files_without_hits:
        # pylint: disable=comparison-with-itself
        # NaN check: float('nan') != float('nan').
        if xp_result != xp_result:
            if args.files_without_hits:
                print(source_name)
        elif args.files_with_hits and xp_result:
            print(source_name)
        # False is a possible value for xp_result (XPath test).
        elif args.files_without_hits and not xp_result:
            print(source_name)
        return True

    # Result header.
    print_result_header(source_name, xp_result)
    # XPath results.
    print_xp_result(xp_result, el_tree, ns_map, args)
    return True


def main():
    """xp command line script entry point."""
    # Logging to the console.
    setup_logger_console()

    # Command line.
    args = parse_cl()

    # XPath expression.
    if isinstance(args.xpath_expr, bytes):
        # Python 2 Bytestring to Unicode.
        args.xpath_expr = args.xpath_expr.decode("utf-8")
    if not build_xpath(args.xpath_expr):
        sys.exit(60)

    # XPath function and XML parser.
    (xpath_fn, xml_parser) = xp_prepare(args)

    # Use XPath on XML sources.
    extra_new_line = False
    for xml_s in args.xml_sources:
        if extra_new_line:
            print()
        elif not (args.files_with_hits or args.files_without_hits):
            extra_new_line = True
        xpath_on_xml(xml_s, xml_parser, xpath_fn, args)

    if not args.xml_sources:
        # Read from a pipe when no XML source is specified.
        if not sys.stdin.isatty():
            xpath_on_xml(sys.stdin, xml_parser, xpath_fn, args)
        else:
            sys.stderr.write("Error: no XML source specified\n")
            sys.exit(70)



# -*- coding: utf-8 -*-

"""Pretty Print XML."""


from __future__ import print_function

import sys
import os
import io
import errno

# pylint: disable=no-name-in-module
# lxml ElementTree <https://lxml.de/>
from lxml.etree import tostring

# Import my own modules.
from .etree import build_etree


__all__ = ['prettyprint', 'pp_xml']

def _private_pp(etree, syntax=True, xml_declaration=None):
    """Pretty print XML ElementTree with (optional) syntax highlighting.

    https://lxml.de/api.html#serialisation
    https://lxml.de/api/lxml.etree-module.html#tostring
    """
    try:
        #if not sys.stdout.isatty() and os.environ['PYTHONIOENCODING']:
        if os.environ['PYTHONIOENCODING']:
            encoding = os.environ['PYTHONIOENCODING']
        elif sys.stdout.encoding is None:
            encoding = 'utf-8'
            #sys.exit("DEBUG")
        else:
            encoding = sys.stdout.encoding
        # lxml.etree.tostring returns bytes (bytestring).
        # https://docs.python.org/3/library/xml.etree.elementtree.html#xml.etree.ElementTree.tostring
        etree_string = tostring(
            etree, encoding=encoding,
            xml_declaration=xml_declaration, pretty_print=True)

        if syntax:
            # pygments.highlight() will return a Unicode string …
            # https://pygments.org/docs/formatters/
            etree_string = highlight(etree_string, lexer, Terminal256Formatter())
        else:
            # Bytes(string) => Unicode object.
            etree_string = etree_string.decode(encoding)

        # Fix output encoding errors (Python 2) when piping multiple etree_string.
        # E.g.: % ppx Unicode_1.xml Unicode_2.xml | less
        #       % xp -q "//d:OtherIdentifier[position()<=12]/.." Unicode.xml -p | less
        if not sys.stdout.isatty() and sys.stdout.encoding is None:
            # Standard out is buffered.
            sys.stdout.flush()
            sys.stdout = io.open(
                # Do not close sys.stdout file descriptor.
                sys.stdout.fileno(), 'w', encoding=encoding, closefd=False)
            print(etree_string)
            # Restore sys.stdout (reuse file descriptor).
            sys.stdout = sys.__stdout__
        else:
            print(etree_string)
    # Python 2: catch 'IOError: [Errno 32] Broken pipe' (multiple etrees).
    except IOError as e:
        if e.errno != errno.EPIPE:
            sys.stderr.write("IOError: %s [%s]\n" % (e.strerror, e.errno))


try:
    # pylint: disable=wrong-import-position
    from pygments.lexers import get_lexer_by_name
    from pygments.formatters import Terminal256Formatter
    from pygments import highlight
except ImportError:
    # pylint: disable=unused-argument
    def prettyprint(etree, syntax=False, xml_declaration=None):
        """Plain pretty print XML ElementTree (without syntax highlighting)."""
        return _private_pp(etree, syntax=False, xml_declaration=xml_declaration)
else:
    lexer = get_lexer_by_name('xml')
    def prettyprint(etree, syntax=True, xml_declaration=None):
        """Pretty print XML ElementTree with (optional) syntax highlighting."""
        return _private_pp(etree, syntax=syntax, xml_declaration=xml_declaration)


def pp_xml(xml_source, parser=None, syntax=True, xml_declaration=None):
    """Pretty Print XML source.

    xml_source -- XML file, file-like object or URL
    parser -- (optional) XML parser (lxml.etree.XMLParser)
    syntax -- syntax highlighting (or not)
    xml_declaration -- print XML declaration (or not)
    """
    xml_tree = build_etree(xml_source, parser=parser)
    if xml_tree:
        prettyprint(xml_tree, syntax=syntax, xml_declaration=xml_declaration)



# -*- coding: utf-8 -*-

"""ElementTree XML API.

The ElementTree XML API:
    https://docs.python.org/library/xml.etree.elementtree.html

ElementTree Overview:
    https://effbot.org/zone/element-index.htm

The lxml.etree Tutorial
    https://lxml.de/tutorial.html
"""


from logging import getLogger
import sys

# pylint: disable=no-member
# lxml ElementTree <https://lxml.de/>
from lxml import etree


# Module logging initialisation.
logger = getLogger(__name__)


def build_etree(xml_source, parser=None, lenient=True, silent=False):
    """Parse XML source into an ElementTree.

    xml_source -- XML file, file-like object or URL
    parser -- (optional) XML parser (lxml.etree.XMLParser)
    lenient -- log XMLSyntaxError as warnings instead of errors
    silent -- no logging

    Return ElementTree (lxml.etree._ElementTree) on success.
    Return None on error.

    Extensible Markup Language (XML):
        https://www.w3.org/XML/

    The lxml.etree.parse function:
        https://lxml.de/parsing.html
        https://effbot.org/zone/element.htm#reading-and-writing-xml-files
    Parser options (lxml.etree.XMLParser class):
        https://lxml.de/parsing.html#parser-options
        https://effbot.org/elementtree/elementtree-xmlparser.htm
    """
    # XML parser preparation.
    if not parser:
        parser = etree.XMLParser(ns_clean=True)

    try:
        el_tree = etree.parse(xml_source, parser)
    # Catch XML syntax errors.
    #   https://lxml.de/api.html#error-handling-on-exceptions
    # error log copy attached to the exception: global error log of all errors
    # that occurred at the application level.
    except etree.XMLSyntaxError:
        if silent:
            return None
        if lenient:
            xmllogger = logger.warning
        else:
            xmllogger = logger.error
        if xml_source in ('-', sys.stdin):
            name = sys.stdin.name
            xml_type = "object"
        else:
            name = xml_source
            xml_type = "file"
        xmllogger("%s is not a valid XML %s:", name, xml_type)

        # Parsers have an error_log property that lists the errors and warnings
        # of the last parser run.
        #   https://lxml.de/parsing.html#error-log
        for e in parser.error_log:
            # For example: e.level_name: "FATAL", e.domain_name: "PARSER",
            # e.type_name: "ERR_DOCUMENT_EMPTY"
            if e.line == 0:
                logger.error(e.message)
            else:
                xmllogger("line %i, column %i: %s", e.line, e.column, e.message)
        return None
    # Catch EnvironmentError (IOError) exceptions, for example:
    #   "failed to load external entity" (lxml.etree._raiseParseError)
    except EnvironmentError as e:
        logger.error(e)
        return None
    else:
        return el_tree



# -*- coding: utf-8 -*-

"""XSL Transformations.

The Extensible Stylesheet Language Family (XSL):
    https://www.w3.org/Style/XSL/

XSL Transformations (XSLT):
    https://www.w3.org/TR/xslt

XSLT with lxml
    https://lxml.de/xpathxslt.html#xslt
"""


from logging import getLogger
import sys

# pylint: disable=no-member
# lxml ElementTree <https://lxml.de/>
from lxml import etree

# Import my own modules.
from .etree import build_etree


# Module logging initialisation.
logger = getLogger(__name__)


def build_xsl_transform(xslt_source):
    """Parse an XSLT source into an XSL Transformer.

    xslt_source -- XSLT file, file-like object or URL

    Lines with XSLT parse errors are logged as warnings.

    Return XSL Transformer (lxml.etree.XSLT) on success.
    Return None on error.

    The lxml.etree.XSLT class:
        https://lxml.de/api/lxml.etree.XSLT-class.html
    I/O access control in XSLT:
        https://lxml.de/resolvers.html#i-o-access-control-in-xslt
    """
    xslt_etree = build_etree(xslt_source, lenient=False)
    if not xslt_etree:
        return None

    try:
        xsl_transform = etree.XSLT(xslt_etree)
    # Catch XSLT parse errors.
    except etree.XSLTParseError as inst:
        logger.error("XML source '%s' is not a valid XSLT source", xslt_source)
        if not inst.error_log:
            logger.error("XSLTParseError: %s", inst)
        for e in inst.error_log:
            # E.g. e.level_name: "ERROR", e.domain_name: "XSLT", e.type_name: "ERR_OK"
            if e.line == 0:
                logger.error(e.message)
            else:
                logger.warning("line %i, column %i: %s", e.line, e.column, e.message)
        return None
    else:
        return xsl_transform


def etree_transformer(el_tree, transformer, **params):
    """Transform an ElementTree with an XSL Transformer.

    el_tree -- ElementTree (lxml.etree._ElementTree)
    transformer -- XSL Transformer (lxml.etree.XSLT)
    params -- (optional) XSL style sheet parameters:
        https://lxml.de/xpathxslt.html#stylesheet-parameters

    XSLT lines with apply errors are logged as warnings.

    Return lxml.etree._XSLTResultTree object on success.
    Return None on error.
    """
    try:
        if params:
            xslt_result = transformer(el_tree, **params)
        else:
            xslt_result = transformer(el_tree)
    # Catch XSL Transformation errors.
    except etree.XSLTApplyError as inst:
        if not inst.error_log:
            logger.error("XSLTApplyError: %s", inst)
        else:
            logger.error("XSLTApplyError on ElementTree")
        for e in inst.error_log:
            # E.g. e.level_name: "ERROR", e.domain_name: "XSLT", e.type_name: "ERR_OK"
            if e.line == 0:
                logger.error(e.message)
            else:
                logger.warning("line %i, column %i: %s", e.line, e.column, e.message)
        return None
    else:
        return xslt_result


def xml_transformer(xml_source, transformer, parser=None):
    """Transform an XML source with an XSL Transformer.

    xml_source -- XML file, file-like object or URL
    transformer -- XSL Transformer (lxml.etree.XSLT)
    parser -- (optional) XML parser (lxml.etree.XMLParser)

    Return lxml.etree._XSLTResultTree object on success.
    Return None on error.
    """
    el_tree = build_etree(xml_source, parser=parser, lenient=False)
    if not el_tree:
        return None

    xslt_result = etree_transformer(el_tree, transformer)
    if xslt_result:
        return xslt_result

    if xml_source in ('-', sys.stdin):
        name = sys.stdin.name
    else:
        name = xml_source
    logger.error("XSL transformation on '%s' failed", name)
    return None



# -*- coding: utf-8 -*-

"""Pretty Print XML source in human readable form."""


from argparse import ArgumentParser
from sys import stdin, stderr

# pylint: disable=no-name-in-module
# lxml ElementTree <https://lxml.de/>
from lxml.etree import XMLParser

# Import my own modules.
from .. import __version__
from ..log import setup_logger_console
from ..ppxml import pp_xml


def parse_cl():
    """Parse the command line for options and XML sources."""
    parser = ArgumentParser(
        description=__doc__)
    parser.add_argument(
        "-V", "--version", action="version",
        version="%(prog)s " + __version__)
    parser.add_argument(
        "xml_sources", nargs='*',
        metavar='xml_source', help="XML source (file, <stdin>, http://...)")
    parser.add_argument(
        "-n", "--no-syntax", action="store_false", default=True,
        dest="syntax", help="no syntax highlighting")
    parser.add_argument(
        "-o", "--omit-declaration", action="store_false", default=True,
        dest="declaration", help="omit the XML declaration")
    return parser.parse_args()


def main():
    """ppx command line script entry point."""
    # Logging to the console.
    setup_logger_console()

    # Command line.
    args = parse_cl()

    # Initialise XML parser and remove blank text for 'pretty_print' formatting.
    #   https://lxml.de/FAQ.html#parsing-and-serialisation
    parser = XMLParser(remove_blank_text=True)

    # Pretty print XML sources.
    for xml_s in args.xml_sources:
        pp_xml(
            xml_s, parser=parser,
            syntax=args.syntax,
            xml_declaration=args.declaration)

    if not args.xml_sources:
        # Read from a pipe when no XML source is specified.
        if not stdin.isatty():
            pp_xml(
                stdin, parser=parser,
                syntax=args.syntax,
                xml_declaration=args.declaration)
        else:
            stderr.write("Error: no XML source specified\n")



# -*- coding: utf-8 -*-

"""XML Utilities.

XML scripts in Python.

 * ppx: pretty print XML source in human readable form.
 * xp: use XPath expression to select nodes in XML source.
 * validate: validate XML source with XSD or DTD.
 * transform: transform XML source with XSLT.
"""


# Xul version.
__version_info__ = ('2', '4', '1')
__version__ = '.'.join(__version_info__)



# -*- coding: utf-8 -*-

"""Logging.

Python logging URLs:
    https://docs.python.org/library/logging.html
    https://docs.python.org/library/logging.handlers.html
    https://docs.python.org/library/logging.config.html
    https://docs.python.org/howto/logging.html
    https://docs.python.org/howto/logging-cookbook.html
"""


import logging


# Module logging initialisation.
logger = logging.getLogger(__name__)

# Logging level names.
log_levels = {
    'debug':    logging.DEBUG,
    'info':     logging.INFO,
    'warning':  logging.WARNING,
    'error':    logging.ERROR,
    'critical': logging.CRITICAL}


def lvl_name2num(name):
    """Return the numeric value of the `name` logging level.

       Log an error for nonexistent `name` and return `logging.NOTSET`.
            https://docs.python.org/library/logging.html#levels
    """
    try:
        levelno = log_levels[name]
    except KeyError:
        logger.error("Log level not set: '%s' is not a valid level", name)
        logger.error("Fix the log level configuration")
        return logging.NOTSET
    else:
        return levelno


def setup_logger(log_level='debug', propagate=True):
    """Configureer threshold log level van de root logger.

       log_level -- threshold log level
       propagate -- messages doorgeven naar boven [default: ja]

       Log records worden door message handlers verwerkt.
       Met setup_logger_console() wordt de console handler gekoppeld.
    """
    # logger '' staat voor de root logger
    logging.getLogger('').setLevel(lvl_name2num(log_level))
    if not propagate:
        # Er hoeft niet dubbel gelogd te worden
        logging.getLogger('').propagate = False


def customize_handler(handler, level, fmt=None, datefmt=None):
    """Configureer level en formatering van een log message handler.

       handler -- log message handler (StreamHandler, FileHandler ...)
       level -- log level voor de handler
       fmt -- formatering voor LogRecord
            https://docs.python.org/library/logging.html#logrecord-attributes
       datefmt -- datum en tijd formatering

       Geef log message handler terug.
    """
    # Configureer log level
    handler.setLevel(lvl_name2num(level))
    # Formatering aanpassen, indien gewenst.
    if fmt or datefmt:
        handler.setFormatter(logging.Formatter(fmt=fmt, datefmt=datefmt))

    return handler


def setup_logger_console(log_level='info'):
    """Setup de root logger en koppel de console handler.

       log_level -- console log level [default 'info']

       Logging op het console (sys.stderr) voor de command line.

       Geef console handler (StreamHandler) terug.
            https://docs.python.org/library/logging.handlers.html#streamhandler
    """
    # Configureer threshold log level DEBUG voor de root logger (i.p.v. WARNING).
    setup_logger()

    # Configureer de console handler.
    console_handler = customize_handler(logging.StreamHandler(), log_level)
    # Koppel console handler aan de root logger.
    logging.getLogger('').addHandler(console_handler)

    return console_handler



# -*- coding: utf-8 -*-

"""XPath.

XML Path Language
    https://www.w3.org/TR/xpath/

Namespaces in XML 1.0
    https://www.w3.org/TR/xml-names/

XPath with lxml
    https://lxml.de/xpathxslt.html#xpath

The XPath result depends on the XPath expression used.
    https://lxml.de/xpathxslt.html#xpath-return-values
- attributes:               "location/@attribute"
- text nodes:               "location/text()"
- element nodes:            "location"
- comment nodes:            "location/comment()"
- processing instructions:  "location/processing-instruction()"
- boolean:                  "location/text() = 'string'"
- string:                   "namespace-uri(location)"
- float:                    "count(location)"
"""


from logging import getLogger

# pylint: disable=no-name-in-module
# lxml ElementTree <https://lxml.de/>
from lxml.etree import XPath, XPathSyntaxError, XPathEvalError
from lxml.etree import LIBXSLT_COMPILED_VERSION

# Import my own modules.
from .etree import build_etree


# Module logging initialisation.
logger = getLogger(__name__)


def build_xpath(xpath_exp, ns_map=None):
    """Build an lxml.etree.XPath instance from an XPath expression.

    xpath_exp -- XPath expression
    ns_map -- XML namespace (prefix: URI) dictionary

    Uses the lxml XPath class:
        https://lxml.de/xpathxslt.html#the-xpath-class
        https://lxml.de/api/lxml.etree.XPath-class.html
    """
    if not ns_map:
        ns_map = {}
    try:
        xpath_obj = XPath(xpath_exp, namespaces=ns_map)
    # Handle (parsing) errors in XPath expression.
    #   https://lxml.de/xpathxslt.html#error-handling
    except XPathSyntaxError as e:
        logger.error("Invalid XPath '%s'", xpath_exp)
        logger.error("XPathSyntaxError: %s", e)
        return None
    else:
        return xpath_obj


def etree_xpath(el_tree, xpath_obj):
    """Apply XPath instance to an ElementTree.

    el_tree -- ElementTree (lxml.etree._ElementTree)
    xpath_obj -- lxml.etree.XPath instance; see build_xpath()
    """
    try:
        xpath_result = xpath_obj(el_tree)
    # Handle errors in evaluating an XPath expression.
    #   https://lxml.de/xpathxslt.html#error-handling
    except XPathEvalError as e:
        logger.error("Invalid XPath '%s'", xpath_obj)
        logger.error("XPathEvalError: %s", e)
        return None
    # Incorrect EXSLT function call (e.g. number of arguments for re:match).
    except TypeError as e:
        logger.error("Invalid XPath '%s'", xpath_obj)
        logger.error("TypeError: %s", e)
        return None
    else:
        return xpath_result


def call_xpath(xml_source, xpath_obj):
    """Apply lxml.etree.XPath on an XML source.

    xml_source -- XML file, file-like object or URL
    xpath_obj -- lxml.etree.XPath instance; see build_xpath()
    """
    # Parse an XML source into an XML Document Object Model.
    el_tree = build_etree(xml_source, lenient=False)
    if not el_tree:
        # No XML source.
        return None

    xpath_result = etree_xpath(el_tree, xpath_obj)
    if xpath_result is None:
        logger.error("XPath failed on %s", xml_source)
    return xpath_result


def xml_xpath(xml_source, xpath_exp):
    """Apply XPath expression to an XML source.

    xml_source -- XML file, file-like object or URL
    xpath_exp -- XPath expression

    Uses call_xpath()
    """
    xpath_obj = build_xpath(xpath_exp)
    if not xpath_obj:
        return None

    return call_xpath(xml_source, xpath_obj)


def update_ns_map(ns_map, elm, none_prefix='default'):
    """Update XPath namespace prefix mapping with element namespaces.

    ns_map -- an XML namespace prefix mapping
    elm -- element with namespaces
    none_prefix -- prefix for the default namespace in XPath

    Element namespaces:
    - xmlns, default namespace (None prefix) URI: elm.nsmap[None]
    - xmlns:prefix, namespace URI: elm.nsmap[elm.prefix]

    Remarks:
     * XPath does not have a notion of a default namespace.
       The empty namespace prefix is not supported in XPath (TypeError).
     * No protection against namespace prefix collisions.
       First occurrence (ns_map) wins.

    Namespaces and prefixes:
        https://lxml.de/xpathxslt.html#namespaces-and-prefixes
    """
    for key in elm.nsmap:
        if not key:
            # XPath prefix for element default namespace.
            if not none_prefix in ns_map:
                ns_map[none_prefix] = elm.nsmap[key]
        elif not key in ns_map:
            # Protect the XPath default namespace prefix.
            if not key == none_prefix:
                ns_map[key] = elm.nsmap[key]


def namespaces(el_tree, exslt=False, none_prefix='default'):
    """Collect all XML namespaces (xmlns) in ElementTree.

    el_tree -- ElementTree (lxml.etree._ElementTree)
    exslt -- add EXSLT XML namespace prefixes (libxslt 1.1.25 and newer)
    none_prefix -- prefix for the default namespace in XPath

    Return XML namespaces (xmlns) 'prefix: URI' dict.

    Namespaces.
        https://lxml.de/tutorial.html#namespaces
    """
    if exslt:
        if LIBXSLT_COMPILED_VERSION < (1, 1, 25):
            logger.warning(
                "EXSLT requires libxslt 1.1.25 or higher. " +
                "lxml is compiled against libxslt %s",
                '.'.join(str(n) for n in LIBXSLT_COMPILED_VERSION))
        # EXSLT <http://exslt.org/>
        ns_map = {
            'date': "http://exslt.org/dates-and-times",
            'dyn': "http://exslt.org/dynamic",
            'exsl': "http://exslt.org/common",
            'func': "http://exslt.org/functions",
            'math': "http://exslt.org/math",
            'random': "http://exslt.org/random",
            're': "http://exslt.org/regular-expressions",
            'set': "http://exslt.org/sets",
            'str': "http://exslt.org/strings"}
    else:
        ns_map = {}

    # Collect XML namespaces (xmlns) in all elements.
    for elm in el_tree.iter('*'):
        if elm.nsmap:
            update_ns_map(ns_map, elm, none_prefix=none_prefix)

    return ns_map



# -*- coding: utf-8 -*-

"""Validate XML source with XSD, DTD or RELAX NG."""


from __future__ import print_function

from argparse import ArgumentParser
import sys

# Import my own modules.
from .. import __version__
from ..log import setup_logger_console
from ..validate import build_xml_schema, build_dtd, build_relaxng
from ..validate import validate_xml


def parse_cl():
    """Parse the command line for options and XML sources."""
    parser = ArgumentParser(description=__doc__)
    parser.add_argument(
        "-V", "--version", action="version",
        version="%(prog)s " + __version__)
    lang_group = parser.add_mutually_exclusive_group(required=True)
    lang_group.add_argument(
        "-x", "--xsd",
        action="store", dest="xsd_source",
        help="XML Schema Definition (XSD) source")
    lang_group.add_argument(
        "-d", "--dtd",
        action="store", dest="dtd_source",
        help="Document Type Definition (DTD) source")
    lang_group.add_argument(
        "-r", "--relaxng",
        action="store", dest="relaxng_source",
        help="RELAX NG source")
    file_group = parser.add_mutually_exclusive_group(required=False)
    file_group.add_argument(
        "-f", "-l", "--validated-files",
        action="store_true", default=False, dest="validated_files",
        help="only the names of validated XML files are written to standard output")
    file_group.add_argument(
        "-F", "-L", "--invalidated-files",
        action="store_true", default=False, dest="invalidated_files",
        help="only the names of invalidated XML files are written to standard output")
    parser.add_argument(
        "xml_sources", nargs='*',
        metavar='xml_source', help="XML source (file, <stdin>, http://...)")
    return parser.parse_args()

def apply_validator(xml_source, validator, args):
    """Apply XML validator on an XML source."""
    if args.validated_files or args.invalidated_files:
        valid = validate_xml(xml_source, validator, silent=True)
        if (valid and args.validated_files) or (not valid and args.invalidated_files):
            if xml_source in ('-', sys.stdin):
                # <stdin>.
                print(sys.stdin.name)
            else:
                print(xml_source)
    else:
        validate_xml(xml_source, validator)

def main():
    """validate command line script entry point."""
    # Logging to the console.
    setup_logger_console()

    # Command line.
    args = parse_cl()

    # XSD or DTD Validator?
    if args.xsd_source:
        validator = build_xml_schema(args.xsd_source)
    elif args.dtd_source:
        validator = build_dtd(args.dtd_source)
    elif args.relaxng_source:
        validator = build_relaxng(args.relaxng_source)
    else:
        validator = None
    # Check validator.
    if not validator:
        sys.exit(60)

    # Validate XML sources.
    for xml_s in args.xml_sources:
        apply_validator(xml_s, validator, args)

    if not args.xml_sources:
        if not sys.stdin.isatty():
            # Read from a pipe when no XML source is specified.
            apply_validator(sys.stdin, validator, args)
        else:
            sys.stderr.write("Error: no XML source specified\n")
            sys.exit(70)



# -*- coding: utf-8 -*-

"""Transform XML source with XSLT."""


from __future__ import print_function

from argparse import ArgumentParser, FileType
import sys
import errno

# pylint: disable=no-name-in-module
# lxml ElementTree <https://lxml.de/>
from lxml.etree import XMLParser, tostring

# Import my own modules.
from .. import __version__
from ..log import setup_logger_console
from ..xsl import build_xsl_transform, xml_transformer


def parse_cl():
    """Parse the command line for options, XSLT source and XML sources."""
    parser = ArgumentParser(
        description=__doc__)
    parser.add_argument(
        "-V", "--version", action="version",
        version="%(prog)s " + __version__)
    parser.add_argument("xslt_source", help="XSLT source (file, http://...)")
    parser.add_argument(
        "xml_source", nargs='?', default=sys.stdin, type=FileType('r'),
        help="XML source (file, <stdin>, http://...)")
    output_group = parser.add_mutually_exclusive_group(required=False)
    output_group.add_argument(
        "-x", "--xsl-output", action="store_true", default=False,
        dest="xsl_output", help="honor xsl:output")
    output_group.add_argument(
        "-o", "--omit-declaration", action="store_false", default=True,
        dest="declaration", help="omit the XML declaration")
    parser.add_argument(
        "-f", "--file", dest="file", help="save result to file")
    return parser.parse_args()


def print_result(result):
    """Print transformation result."""
    try:
        print(result)
    # io.TextIOWrapper catches Python 3 BrokenPipeError.
    except IOError as e:
        # Python 2: catch 'IOError: [Errno 32] Broken pipe'.
        if e.errno != errno.EPIPE:
            sys.stderr.write("IOError: %s [%s]\n" % (e.strerror, e.errno))
    except LookupError as e:
        # LookupError: unknown encoding: UCS-4.
        sys.stderr.write("LookupError (XSLT result): %s\n" % e)


def save_to_file(result, target_file):
    """Save transformation result to file."""
    if sys.version_info[0] == 2:
        # File exists?
        file_mode = "wb"
    else:
        file_mode = "bx"
    try:
        with open(target_file, file_mode) as file_object:
            file_object.write(result)
    except EnvironmentError as e:
        sys.stderr.write("Saving result to %s failed: %s\n" % (
            target_file, e.strerror))
        sys.exit(80)


def output_xslt(xml_source, transformer, parser, args):
    """Print or save the result of an XSL Transformation.

    xml_source -- XML file, file-like object or URL
    transformer -- XSL Transformer (lxml.etree.XSLT)
    parser -- XML parser (lxml.etree.XMLParser)
    args -- Command-line arguments
    """
    result = xml_transformer(xml_source, transformer, parser)
    if not result:
        pass
    elif result.getroot() is None:
        # Result is not an ElementTree.
        print(result)

    # https://lxml.de/xpathxslt.html#xslt-result-objects
    # https://lxml.de/api/lxml.etree._XSLTResultTree-class.html
    #   _XSLTResultTree (./src/lxml/xslt.pxi):
    elif args.xsl_output:
        # https://lxml.de/api/lxml.etree.XSLT-class.html
        # XSLT.tostring(). Deprecated: use str(result_tree) instead.
        #
        # Python 2: str(_XSLTResultTree) == bytes(_XSLTResultTree).
        #
        # Python 3: str(_XSLTResultTree) != bytes(_XSLTResultTree).
        if args.file:
            save_to_file(result, args.file)
        else:
            # Standard output: sys.stdout.encoding.
            # Document labelled UTF-16 but has UTF-8 content:
            #   str(result, result.docinfo.encoding) ==
            #       bytes(result).decode(result.docinfo.encoding)
            print_result(result)

    # https://lxml.de/parsing.html#serialising-to-unicode-strings
    # For normal byte encodings, the tostring() function automatically adds
    # a declaration as needed that reflects the encoding of the returned string.
    else:
        etree_result = tostring(
            result, encoding=result.docinfo.encoding,
            xml_declaration=args.declaration)
        if args.file:
            save_to_file(etree_result, args.file)
        else:
            # lxml.etree.tostring returns bytes (bytestring).
            if not isinstance(etree_result, str):
                # Bytes => unicode string (Python 3).
                etree_result = etree_result.decode(result.docinfo.encoding)
            print_result(etree_result)


def main():
    """transform command line script entry point."""
    # Logging to the console.
    setup_logger_console()

    # Command line.
    args = parse_cl()

    # Check XSLT source.
    if args:
        # Build an XSL Transformer from an XSLT source.
        transformer = build_xsl_transform(args.xslt_source)
        if not transformer:
            sys.stderr.write('Invalid XSLT source specified\n')
            sys.exit(60)
    else:
        sys.stderr.write('No XSLT source specified\n')
        sys.exit(50)

    # Initialise XML parser.
    parser = XMLParser()
    # Transform XML source with XSL Transformer.
    output_xslt(args.xml_source, transformer, parser, args)



# -*- coding: utf-8 -*-

"""Setup script for building XML Utilities."""

import codecs
from os import path
from setuptools import setup, find_packages
from xul import __version__

here = path.abspath(path.dirname(__file__))

def long_description():
    """Get the long description from the README.rst file."""
    with codecs.open(path.join(here, 'README.rst'), encoding='utf-8') as f:
        return f.read()

setup(
    name="Xul",
    version=__version__,

    packages=find_packages(),
    install_requires=["lxml>=2.0"],
    extras_require={
        'syntax': ["Pygments>=2.0"]
    },

    author='Peter Adrichem',
    author_email='Peter.Adrichem@gmail.com',
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: Implementation :: CPython",
        "Topic :: Text Processing :: Markup :: XML",
        "Topic :: Utilities"
    ],
    description="XML (XPath, XSD, XSLT) Utilities",
    keywords="xml, xpath, xslt, xsd, dtd, xml schema, relax ng, rng",
    license='MIT',
    long_description=long_description(),
    long_description_content_type='text/x-rst',
    platforms='CPython',
    python_requires=">=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, !=3.4.*",
    url='https://xul.readthedocs.io/',
    project_urls={
        'Documentation': 'https://xul.readthedocs.io/',
        'Changelog': 'https://xul.readthedocs.io/en/latest/changelog.html',
        'Source': 'https://bitbucket.org/peteradrichem/xul'
    },
    entry_points={
        'console_scripts': [
            'transform = xul.cmd.transform:main',
            'ppx = xul.cmd.ppx:main',
            'xp = xul.cmd.xp:main',
            'validate = xul.cmd.validate:main'
        ]
    }
)
